import { APIClient } from "../entities/APIClient";
import { ConversationMessage } from "../entities/ConversationMessage";
import { Comment } from "../entities/Comment";

/**
 * Events manager
 * 
 * @author Pierre Hubert
 */

// When a user sign out
export interface DestroyedLoginTokensEvent {
	userID: number,
	client: APIClient
}

// When some user might have an updated number of notifications
export interface UpdatedNotificationsNumberEvent {
	usersID: number[]
}

// When some users have an updated number of unread conversations
export interface UpdateNumberUnreadConversationsEvent {
	usersID: number[]
}

// When a new conversatino message is sent
export interface SentNewConversationMessageEvent {
	msg: ConversationMessage
}

// When a comment is created
export interface CommentCreatedEvent {
	comment: Comment
}

/**
 * Global map of all possible events
 */
export interface EventsMap {
	"destroyed_login_tokens": DestroyedLoginTokensEvent,
	"updated_number_notifications": UpdatedNotificationsNumberEvent,
	"updated_number_unread_conversations": UpdateNumberUnreadConversationsEvent,
	"sent_conversation_message": SentNewConversationMessageEvent,
	"comment_created": CommentCreatedEvent
}

export class EventsHelper {

	/**
	 * The list of events
	 */
	private static EventsList: Map<keyof EventsMap, Array<(ev: any) => void | Promise<void>>> = new Map()

	/**
	 * Listen to a new event
	 * 
	 * @param name The name of the event to listen to
	 * @param listener Target listener
	 */
	public static Listen<K extends keyof EventsMap>(name: K, listener: (ev: EventsMap[K]) => void | Promise<void>) {
		if(!this.EventsList[<any>name])
			this.EventsList[<any>(name)] = [];

		this.EventsList[<any>(name)].push(listener);
		
	}

	/**
	 * Push a new event
	 * 
	 * Note : It is not mandatory to listen to this method, as
	 * we will try / catch all the events
	 * 
	 * @param name The name of the event
	 * @param args Argumens to pass to the event
	 */
	public static async Emit<K extends keyof EventsMap>(name: K, args: EventsMap[K]) {
		
		if(!this.EventsList[<any>name])
			return;

		for(const entry of this.EventsList[<any>name]) {
			try {
				await entry(args);
			}
			catch(e) {
				console.error("An error occured while propagating an event ("+name+")!", e, args);
			}
		}
	}

	
}